
#target tap

; tape block sync bytes:
headerflag:     = 0
dataflag:       = 0xff

; some Basic tokens:
tCLEAR		= $FD		; token CLEAR
tLOAD		= $EF		; token LOAD
tCODE		= $AF		; token CODE
tPRINT		= $F5		; token PRINT
tRANDOMIZE	= $F9		; token RANDOMIZE
tUSR		= $C0		; token USR

; memory layout:
screen_start	equ 0x4000	; ZXSP screen pixels
attr_start	equ 0x5800	; ZXSP screen attributes
attr_end	equ 0x5B00	; ZXSP printer buffer
slow_ram	equ 0x5B00	; start of usable slow ram
fast_ram	equ 0x8000	; start of fast ram
ram_end 	equ $10000

load_address	equ $10000 - code_size 	; load code block here (will be relocated)
code_address 	equ attr_end		; where it is decompressed/moved to
start_address	equ start		; mcode start address

; settings:
stack_size	equ 512		; VM return + VM variables stack size



; ---------------------------------------------------
;	a Basic Loader:
; ---------------------------------------------------

#code PROG_HEADER,0,17,headerflag
	db  0			; Indicates a Basic program
	db  "mloader   "	; the block name, 10 bytes long
	dw  variables_end-0	; length of block = length of basic program plus variables
	dw  10		    	; line number for auto-start, 0x8000 if none
	dw  program_end-0	; length of the basic program without variables

#code PROG_DATA,0,*,dataflag

  ; 10 CLEAR load_address-1
        db  0,10		; line number
        db  end10-($+1)		; line length
        db  0			; statement number
        db  tCLEAR		; token CLEAR
        db  "0", $0e0000 \ dw load_address-1 \ db 0   ; number: ascii & internal format
end10:	db  $0d			; line end marker

  ; 20 LOAD "" CODE
        db  0,20		; line number
        db  end20-($+1)		; line length
        db  0			; statement number
        db  tLOAD, '""', tCODE	; token LOAD, 2 quotes, token CODE
end20:	db  $0d			; line end marker

  ; 30 RANDOMIZE USR start_address
        db  0,30		; line number
        db  end30-($+1)		; line length
        db  0			; statement number
        db  tRANDOMIZE, tUSR	; token RANDOMIZE, token USR
        db  "0", $0e0000 \ dw start_address \ db 0   ; number: ascii & internal format
end30:	db  $0d			; line end marker

program_end:

	; ZX Spectrum Basic variables

variables_end:



; ---------------------------------------------------
;	The machine code block:
; ---------------------------------------------------

#code CODE_HEADER, 0, 17, headerflag
	db  3			; Indicates binary data
	db  "mcode     "	; the block name, 10 bytes long
	dw  code_size		; length of data block which follows
	dw  load_address	; loading location for the data
	dw  0       		; unused

#code CODE_DATA, 0, 0, dataflag

; Machine Code and Data:
; note: all following #code blocks are appended to #code CODE_DATA



; ---------------------------------------------------
;	MEMORY LAYOUT
; ---------------------------------------------------

; we put everything in #code segments
; => every byte is actually stored in the tape
;    but we plan to use zx7 compression (TODO)

; slow ram:
#code SLOW_DATA, code_address	; Uncritical/interrupt zeroed variables
#code SLOW_GSDATA		; Uncritical/interrupt initialized variables
#code SLOW_CODE 		; Uncritical/interrupt machine code
#code SLOW_CONST 		; Uncritical/interrupt const data, e.g. compressed levels, music

; fast ram:
#code VCODE			; All virtual code (except boot code): Main Program, Libraries
#code VMCORE			; VM core
#code CODE			; Machine code
#code CONST			; Const data
#code GSDATA			; Initialized variables
#code DATA			; Zeroed variables
#code VMSTACKS, *, stack_size 	; VM return stack + VM variables stack
#code GSINIT	 		; Boot code: mcode + vcode
#code ZX7			; the only uncompressed #code segment
#data HEAP, GSINIT, ram_end-$$	; Managed memory

#compress SLOW_DATA to GSINIT



; ---------------------------------------------------
;	INTERRUPT HANDLER
; ---------------------------------------------------

interrupt_handler:: equ irpt_scan_keyboard	; lib-z80/zxsp_keyboard.s



; ---------------------------------------------------
;	BOOT CODE
; ---------------------------------------------------

; ________________________________
#code ZX7
#assert	load_address >= code_address + SLOW_DATA_to_GSINIT_cgain + SLOW_DATA_to_GSINIT_cdelta
.phase 	load_address + SLOW_DATA_to_GSINIT_csize
start::	di
	ld	sp, initv		; -> ret will jump to gsinit
	ld	hl, load_address
	ld	de, code_address
#include "lib-z80/decompress_zx7_standard.s"


; ________________________________
#code VMSTACKS
; setup the init vector: this is used as return address by the zx7 decompressor
; it is placed as the last return address on the machine stack.
	ds	stack_size - 2
initv:	dw	gsinit			
	

; ________________________________
#code GSINIT
; wrap GSINIT in a 'used' memory block:
	db	1			; ref_cnt
	dw	GSINIT_size -3 -3	; block size = GSINIT size - 2 block headers

gsinit:

#if 1 ; test code
	ld	a,2		; set print channel to upper Screen
	call	$1601
	ld	hl,2$
1$:	ld	a,(hl++)
	and	a
	ret	z
	rst	2
	jr	1$
2$:	dm	13, "Hello World!", 13, 0
#endif

; --- init z80 stack ---
	
	; nothing to do: 
	; already set up by 'ld sp,initv' for zx7!

; --- init heap (GSINIT is first block of heap!) ---
	
	; TODO
	; probably not much to do

; --- #include mcode with GSINIT chunks ---

#include "core-interrupt-zx48.s"
#include "lib-z80/zxsp_keyboard.s"

; --- setup interrupt ---
; adjust position of this code so that the gap before 'intv' is small

#code SLOW_DATA
	ds 	255 - lo($$)		; low byte of address of interrupt_vector must be 255
intv:	dw 	vm_interrupt_handler	; interrupt vector
#code GSINIT
	ld	a, hi(intv)
	ld	i,a

; --- #include more mcode with GSINIT chunks ---

#include "core-macros.s"

; --- switch to virtual mode ---

#code GSINIT
	ld	bc, VMSTACKS	 ; VM Rstack: bottom up
	exx
	ld	bc, VMSTACKS_end ; VM Vstack: top down

	ld	sp, $+4		 ; VM PC
	ret
	
; --- #include vcode with GSINIT chunks ---

#cpath 	 "vcc"	
#cflags  -cpu=z80-rop -DDEBUG -o4 -v -c
;#include "main.cc"

; --- release the memory block which wraps the GSINIT code ---

#code GSINIT
	dw	NATIVE
	xor	a
	ld	(GSINIT),a	; -> ref_cnt of block with boot code

; --- enable interrupt ---

	ei	

; --- jump to function main() (which must not return) ---

	;ld	bc, VMSTACKS	 ; VM Rstack	still valid if BC wasn't used
	;exx
	;ld	bc, VMSTACKS_end ; VM Vstack	still valid if B'C' wasn't used
	
	ld	sp, main+6	 ; assumes: sizeof(p_enter) = 6
	ret



; ===========================================
;    resolve missing labels from VM engine
; ===========================================
	
#code VMCORE
;#include library "../lib-Z80-Rop"
;#include library "lib-z80"



;TODO: preliminary	
p_enter::
	pushr_sp		;61	push current VM PC on rstack
	ld	sp,hl		;4	load VM PC
	ret			;10	next
RETURN::		; 54cc
	popr_sp		;44
	ret		;10
NATIVE::
	ld	hl,0
	add	hl,sp		; HL = VM PC
	jp	hl		; PC = VM PC
main::	p_enter
	dw	RETURN



; ===========================================
;    fixings
; ===========================================

#code GSINIT
; setup a 'free memory block' header for the remaining ram:
	db	0		; ref_cnt 0 = free
	dw	ram_end - $	; block size


#code SLOW_CONST
; OPTIONAL: raise VCODE into fast ram
	.org   	max($$,$8000)	
; assert requirements by VCODE checksum method:
; 1. mutable data before VCODE must not end in a page overlapping with VCODE:
; 2. VCODE must be word-aligned:
	.align	hi(SLOW_GSDATA_end-1) == hi($$) ? 256 : 2


#code VMCORE
; raise VMCORE into fast ram
	.org   	max($$,$8000)	


#code CONST
; assert requirements by VCODE checksum method:
; 3. mutable data behind VCODE must not start in a page overlapping with VCODE:
	.align	hi($$) == hi(VCODE_end-1) ? 256 : 1
	
#code DATA
; align VMSTACKS to even address
	.align 	2



; ===========================================
;    final calculations
; ===========================================

; allocation in final program (without ZX7 and GSINIT):
mcode_size equ SLOW_CODE_size + CODE_size + VMCORE_size
vcode_size equ VCODE_size
data_size  equ SLOW_GSDATA_size + GSDATA_size + SLOW_CONST_size + CONST_size + SLOW_DATA_size + DATA_size
stack_size equ stack_size
#assert SLOW_DATA_to_GSINIT_size == mcode_size + vcode_size + data_size + stack_size + GSINIT_size

; size of code block on tape:
code_size  equ SLOW_DATA_to_GSINIT_csize + ZX7_size

; size of tape blocks in .tap file:
BLK1SZ	   equ 4 + PROG_HEADER_size
BLK2SZ	   equ 4 + PROG_DATA_size 
BLK3SZ	   equ 4 + CODE_HEADER_size
BLK4SZ	   equ 4 + code_size

; allocation in slow and fast ram in final program (without ZX7 and GSINIT):
slowram_size equ SLOW_DATA_size + SLOW_GSDATA_size + SLOW_CODE_size + SLOW_CONST_size 
fastram_size equ VCODE_size + VMCORE_size + CODE_size + CONST_size + GSDATA_size + DATA_size + VMSTACKS_size
#assert SLOW_DATA_to_GSINIT_size == slowram_size + fastram_size +GSINIT_size

; size of .tap file:
tape_size 		equ BLK1SZ + BLK2SZ + BLK3SZ + BLK4SZ
uncompressed_tape_size	equ BLK1SZ + BLK2SZ + BLK3SZ + 4 + slowram_size + fastram_size + ZX7_size + GSINIT_size
#assert uncompressed_tape_size == tape_size + SLOW_DATA_to_GSINIT_cgain




